/** Controller class for GUI rendering windows.
 *
 *  @author Julian Strohmayer 01426125 <e1426125@student.tuwien.ac.at>
 *  @author Patrick Gantner 01576033 <e1576033@student.tuwien.ac.at>
 *  @date 05.06.2018
 */

#ifndef GL_RENDER_WIDGET_H
#define GL_RENDER_WIDGET_H

#include <QOpenGLWidget>
#include <QOpenGLFunctions>

#include <QOpenGLBuffer>
#include <QOpenGLVertexArrayObject>

#include <QOpenGLShader>
#include <QOpenGLTexture>

#include <QOpenGLDebugLogger>

#include "Renderable.h"
#include <QColor>

#include <memory>
#include <vector>

/**
 * @brief The GLRenderWidget class implementing additional OpenGL capabilities on a QWidget.
 */
class GLRenderWidget : public QOpenGLWidget, protected QOpenGLFunctions
{

public:
    GLRenderWidget(QWidget *parent);
    ~GLRenderWidget();

    /**
     * @brief setShader sets the shader to be used for rendering.
     * @param vert location of vertex shader file
     * @param frag location of fragment shader file
     */
    void setShader(const std::string& vert, const std::string& frag);

    /**
     * @brief setRenderObject sets vertex positions and indices of the rendered quad.
     * @param data vertex positions and indices
     */
    void setRenderObject(RenderData& data);

    /**
     * @brief setVolumeTexture sets the volume that should be rendered in this window.
     * @param volume3Dtex volume as 3D texture
     */
    void setVolumeTexture(QOpenGLTexture* volume3Dtex);

    /**
     * @brief setRegion sets the transfer function parameters for this rendering window.
     * @param sep region seperator value (0,1)
     * @param c1 color of the lower region (less dense)
     * @param c2 color of the upper region (more dense)
     */
    void setRegion(float sep, QColor c1, QColor c2);

    /**
     * @brief setOrientation sets the orientation parameter for this rendering window.
     * @param ori Direction number (0..top, 1..bottom, 2..left, 3..right, 4..front, 5..back)
     */
    void setOrientation(int ori);

protected:
    QOpenGLShaderProgram* shader;
    RenderData objectData;

    QOpenGLVertexArrayObject vao;
    QOpenGLBuffer vbo;
    QOpenGLBuffer indices;

    /**
     * @brief initializeGL performs openGL setup.
     */
    void initializeGL();
    /**
     * @brief resizeGL handles resize events
     * @param width openGL viewport width
     * @param height openGL viewport height
     */
    void resizeGL(int width, int height);

    /**
     * @brief paintGL passes unfiroms to shader program and renders volume into GUI window.
     */
    void paintGL();

    QOpenGLDebugLogger* logger;

private:
    /**
     * @brief updateVBOs handles VBO creation/destruction
     */
    void updateVBOs();

    bool hasData;
    bool hasShader;
    bool hasRegionSeperatorValueAndColors;
    bool dirty;
    size_t numVerts;

    QOpenGLTexture* volume;

    // Transfer function parameters
    int orientation;
    float regionSeperator;
    QColor regionColor1;
    QColor regionColor2;

};

#endif
